const ALL_SOUND_OFF_EVENT_DATA = Array.from({ length: 16 }, (_, index) => new Uint8Array([176 + index, 120, 0]));
export class MidiPlayer {
    constructor({ encodeMidiMessage, filterMidiMessage, json, midiFileSlicer, midiOutput, startScheduler }) {
        this._encodeMidiMessage = encodeMidiMessage;
        this._filterMidiMessage = filterMidiMessage;
        this._json = json;
        this._midiFileSlicer = midiFileSlicer;
        this._midiOutput = midiOutput;
        this._startScheduler = startScheduler;
        this._state = null;
    }
    play() {
        if (this._state !== null) {
            throw new Error('The player is currently playing.');
        }
        return new Promise((resolve) => {
            const stopScheduler = this._startScheduler(({ end, start }) => {
                if (this._state === null) {
                    this._state = { endedTracks: 0, offset: start, resolve, stopScheduler: null };
                }
                this._schedule(start, end, this._state);
            });
            if (this._state === null) {
                stopScheduler();
            }
            else {
                this._state.stopScheduler = stopScheduler;
            }
        });
    }
    stop() {
        var _a, _b;
        if (this._state === null) {
            throw new Error('The player is already stopped.');
        }
        // Bug #1: Chrome does not yet implement the clear() method.
        (_b = (_a = this._midiOutput).clear) === null || _b === void 0 ? void 0 : _b.call(_a);
        ALL_SOUND_OFF_EVENT_DATA.forEach((data) => this._midiOutput.send(data));
        this._stop(this._state);
    }
    _schedule(start, end, state) {
        const events = this._midiFileSlicer.slice(start - state.offset, end - state.offset);
        events
            .filter(({ event }) => this._filterMidiMessage(event))
            .forEach(({ event, time }) => this._midiOutput.send(this._encodeMidiMessage(event), start + time));
        const endedTracks = events.filter(({ event }) => MidiPlayer._isEndOfTrack(event)).length;
        state.endedTracks += endedTracks;
        if (state.endedTracks === this._json.tracks.length) {
            this._stop(state);
        }
    }
    _stop(state) {
        const { resolve, stopScheduler } = state;
        stopScheduler === null || stopScheduler === void 0 ? void 0 : stopScheduler();
        this._state = null;
        resolve();
    }
    static _isEndOfTrack(event) {
        return 'endOfTrack' in event;
    }
}
//# sourceMappingURL=midi-player.js.map